/* Copyright (C) 2023 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergei Zharko [committer] */

/// \file   CbmQaTask.h
/// \brief  A base class for CBM QA task logic
/// \author S.Zharko <s.zharko@gsi.de>
/// \data   12.01.2023


#pragma once

#include "CbmEvent.h"
#include "CbmQaIO.h"
#include "CbmQaTable.h"
#include "FairTask.h"
#include "Logger.h"
#include "TCanvas.h"
#include "TEfficiency.h"
#include "TFolder.h"
#include "TH1.h"
#include "TH2.h"
#include "TH3.h"
#include "TParameter.h"
#include "TProfile.h"
#include "TProfile2D.h"
#include "TProfile3D.h"
#include "TROOT.h"

#include <algorithm>
#include <map>
#include <regex>
#include <string_view>
#include <tuple>
#include <type_traits>

#include <yaml-cpp/yaml.h>

class CbmEvent;
class TClonesArray;

/// Class CbmQaTask is to be inherited with a particular QA-task. It provides mechanisms for storage and management
/// of QA canvases and histograms management
class CbmQaTask : public FairTask, public CbmQaIO {
 public:
  /// Constructor from parameters
  /// \param  name     Name of the task
  /// \param  verbose  Verbose level
  /// \param  isMCUsed Flag: true - MC information is used, false - only reconstructed data QA is processed
  CbmQaTask(const char* name, int verbose, bool isMCUsed);

  /// @brief Default constructor
  CbmQaTask() = delete;  // TODO: Let's see, what can happen, if one deletes default constructor

  /// @brief Destructor
  virtual ~CbmQaTask() = default;

  /// @brief Copy constructor
  CbmQaTask(const CbmQaTask&) = delete;

  /// @brief Move constructor
  CbmQaTask(CbmQaTask&&) = delete;

  /// @brief Copy assignment operator
  CbmQaTask& operator=(const CbmQaTask&) = delete;

  /// @brief Move assignment operator
  CbmQaTask& operator=(CbmQaTask&&) = delete;

  /// @brief Gets name of the setup
  const std::string& GetSetupName() const { return fsSetupName; }

  /// @brief Returns flag, whether MC information is used or not in the task
  bool IsMCUsed() const { return fbUseMC; }

  /// @brief Sets events suppression flag
  ///
  /// By default the QA task runs on the reconstructed events (CbmEvent objects), if the tree of the objects is
  /// presented in the input. This flag disables the events-based routine, so the QA is executed over the whole
  /// Time-slice.
  void SetProcessFullTs(bool bProcessFullTs) { fbProcessFullTs = bProcessFullTs; }

  /// FairTask: Task initialization in the beginning of the run
  InitStatus Init() override;

  /// FairTask: Task reinitialization
  InitStatus ReInit() override;

  /// FairTask: Defines action of the task in the event/TS
  void Exec(Option_t* /*option*/) override;

  /// FairTask: Defines action of the task in the end of run
  void Finish() override;

  /// @brief Sets name of the setup
  void SetSetupName(const char* setup) { fsSetupName = setup; }

  // *****************************************************
  // ** Functions accessible inside the derived classes **
  // *****************************************************

  /// \brief  Method to check, if the QA results are acceptable
  /// \return Map of checks: key - name of the check, value - result of the check
  virtual std::map<std::string, bool> Check() = 0;

 protected:
  /// De-initialize the task
  virtual void DeInit() {}

  /// Initializes data branches
  virtual InitStatus InitDataBranches() { return kSUCCESS; }

  /// Initializes histograms
  virtual InitStatus InitHistograms() { return kSUCCESS; }

  /// Initializes canvases
  virtual InitStatus InitCanvases() { return kSUCCESS; }

  /// Initializes event / time-slice
  virtual InitStatus InitTimeSlice() { return kSUCCESS; }

  /// Method to fill histograms per event or time-slice
  virtual void FillHistograms() {}

  /// Get current event number
  int GetEventNumber() const { return fNofEvents.GetVal(); }

  // ***********************
  // ** Utility functions **
  // ***********************

  /// Checks range of variable
  /// \param name  Name of the variable
  /// \param var   Variable to check
  /// \param lo    Lower limit of the variable
  /// \param hi    Upper limit of the variable
  /// \return  False, if variable exits the range
  template<typename T>
  bool CheckRange(std::string_view name, T var, T lo, T hi) const;

  /// Checks range of variable
  /// \param name    Name of the variable
  /// \param var     Variable to check
  /// \param varErr  Standard Error of the variable
  /// \param lo      Lower limit of the variable
  /// \param hi      Upper limit of the variable
  /// \return        False, if variable exits the range
  template<typename T>
  bool CheckRange(std::string_view name, T var, T varErr, T lo, T hi) const;

  /// Checks ranges for mean and standard deviation
  /// \return  False, if variable exits the range
  bool CheckRange(TH1* h, double meanMax, double rmsMin, double rmsMax);

  /// @brief Puts setup title on the canvas
  void PutSetupNameOnPad(double xMin, double yMin, double xMax, double yMax);

  /// @brief Gets pointer to current event
  CbmEvent* GetCurrentEvent() { return fpCurrentEvent; }

 private:
  /// @brief De-initializes this task
  void DeInitBase();

  TClonesArray* fpBrEvents = nullptr;  ///< Pointer to CbmEvent branch
  CbmEvent* fpCurrentEvent = nullptr;  ///< Pointer to the current event
  std::string fsSetupName = "";  ///< Name of the setup (to draw on the canvases)
  TParameter<int> fNofEvents{"nEvents", 0};  ///< Number of processed events
  bool fbUseMC         = false;              ///< Flag, if MC is used
  bool fbProcessFullTs = false;              ///< If true, routine runs on the full TS even if the CbmEvent branch is in


  ClassDefOverride(CbmQaTask, 0);
};


// *************************************************
// ** Inline and template function implementation **
// *************************************************

// ---------------------------------------------------------------------------------------------------------------------
//
template<typename T>
bool CbmQaTask::CheckRange(std::string_view name, T var, T lo, T hi) const
{
  return CheckRange<T>(name, var, 0., lo, hi);
}

// ---------------------------------------------------------------------------------------------------------------------
//
template<typename T>
bool CbmQaTask::CheckRange(std::string_view name, T var, T varErr, T lo, T hi) const
{
  bool ret = true;
  if (var + 3.5 * varErr < lo) {
    LOG(error) << fName << ": " << name << " is found to be under the lower limit (" << var << " +-" << varErr << " < "
               << lo << ')';
    ret = false;
  }
  if (var - 3.5 * varErr > hi) {
    LOG(error) << fName << ": " << name << " is found to be above the upper limit (" << var << " +-" << varErr << " > "
               << hi << ')';
    ret = false;
  }
  return ret;
}
